#!/env/python
# Copyright 2023 Ultima Genomics Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
# DESCRIPTION
#    Add additional feature annotations to featuremap, to be used from single-read SNV qual recalibration
# CHANGELOG in reverse chronological order
import argparse
import sys

from ugbio_core.consts import DEFAULT_FLOW_ORDER

from ugbio_featuremap.featuremap_utils import annotate_featuremap


def parse_args(argv: list[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="annotate featuremap", description=run.__doc__)
    parser.add_argument(
        "-i",
        "--featuremap_path",
        type=str,
        required=True,
        help="input featuremap file",
    )
    parser.add_argument("--ppmSeq_adapter_version", type=str, default=None, help="ppmSeq adapter version")
    parser.add_argument(
        "-o",
        "--output_featuremap",
        type=str,
        required=True,
        help="Path of annotated featuremap file",
    )
    parser.add_argument("-r", "--ref_fasta", type=str, required=True, help="Reference genome fasta file")
    parser.add_argument(
        "--flow_order",
        type=str,
        default=DEFAULT_FLOW_ORDER,
    )
    parser.add_argument(
        "--motif_length_to_annotate",
        type=int,
        default=3,
    )
    parser.add_argument("--max_hmer_length", type=int, default=20)
    parser.add_argument(
        "-@",
        "--process_number",
        type=int,
        default=0,
        help="""Number of processes to use for parallelization.
             If N < 1, use all-available - abs(N) cores. Default 0""",
    )

    return parser.parse_args(argv[1:])


def run(argv):
    """Add additional feature annotations to featuremap"""
    args = parse_args(argv)

    annotate_featuremap(
        input_featuremap=args.featuremap_path,
        output_featuremap=args.output_featuremap,
        ref_fasta=args.ref_fasta,
        ppmseq_adapter_version=args.ppmSeq_adapter_version,
        flow_order=args.flow_order,
        motif_length_to_annotate=args.motif_length_to_annotate,
        max_hmer_length=args.max_hmer_length,
        process_number=args.process_number,
    )


def main():
    run(sys.argv)


if __name__ == "__main__":
    main()



import argparse
import logging
import os
import statistics
import subprocess
import sys
import tempfile
import warnings
from concurrent.futures import ThreadPoolExecutor
from os.path import basename, dirname
from os.path import join as pjoin

import pandas as pd
import pysam
import xgboost
from sklearn.preprocessing import LabelEncoder
from ugbio_core.logger import logger
from ugbio_ppmseq.ppmSeq_consts import HistogramColumnNames

from ugbio_featuremap import featuremap_consensus_utils
from ugbio_featuremap.featuremap_utils import FeatureMapFields

warnings.simplefilter(action="ignore", category=FutureWarning)


default_custom_info_fields = [
    FeatureMapFields.X_SCORE.value,
    FeatureMapFields.X_EDIST.value,
    FeatureMapFields.X_LENGTH.value,
    FeatureMapFields.X_MAPQ.value,
    FeatureMapFields.X_INDEX.value,
    FeatureMapFields.X_FC1.value,
    FeatureMapFields.X_FC2.value,
    FeatureMapFields.MAX_SOFTCLIP_LENGTH.value,
    FeatureMapFields.X_FLAGS.value,
    HistogramColumnNames.STRAND_RATIO_CATEGORY_START.value,
    HistogramColumnNames.STRAND_RATIO_CATEGORY_END.value,
    "ML_QUAL",
    FeatureMapFields.X_RN.value,
    FeatureMapFields.X_CIGAR.value,
    "rq",
    "tm",
    FeatureMapFields.IS_FORWARD.value,
    FeatureMapFields.IS_DUPLICATE.value,
    FeatureMapFields.READ_COUNT.value,
    FeatureMapFields.FILTERED_COUNT.value,
    FeatureMapFields.TRINUC_CONTEXT_WITH_ALT.value,
    FeatureMapFields.HMER_CONTEXT_REF.value,
    FeatureMapFields.HMER_CONTEXT_ALT.value,
    FeatureMapFields.PREV_1.value,
    FeatureMapFields.PREV_2.value,
    FeatureMapFields.PREV_3.value,
    FeatureMapFields.NEXT_1.value,
    FeatureMapFields.NEXT_2.value,
    FeatureMapFields.NEXT_3.value,
    FeatureMapFields.IS_CYCLE_SKIP.value,
    FeatureMapFields.X_QUAL.value,
]

ppm_custom_info_fields = ["st", "et"]

columns_for_mean_aggregation = [
    FeatureMapFields.X_QUAL.value,
    FeatureMapFields.X_SCORE.value,
    FeatureMapFields.X_EDIST.value,
    FeatureMapFields.X_LENGTH.value,
    FeatureMapFields.X_MAPQ.value,
    FeatureMapFields.X_FC1.value,
    FeatureMapFields.X_FC2.value,
    FeatureMapFields.MAX_SOFTCLIP_LENGTH.value,
    FeatureMapFields.X_FLAGS.value,
    "ML_QUAL",
]
columns_for_min_aggregation = [FeatureMapFields.X_QUAL.value, FeatureMapFields.X_INDEX.value]
columns_for_max_aggregation = [FeatureMapFields.X_QUAL.value, FeatureMapFields.X_INDEX.value]
columns_for_fillna = [FeatureMapFields.IS_CYCLE_SKIP.value]
columns_for_st_et_aggregation = ["st", "et"]

added_agg_features = {
    "alt_reads": ["number of supporting reads for the alternative allele", "Integer"],
    "ref_allele": ["reference allele", "String"],
    "alt_allele": ["alternative allele", "String"],
    "X_QUAL_mean": ["mean value of X_QUAL", "Float"],
    "X_SCORE_mean": ["mean value of X_SCORE", "Float"],
    "X_EDIST_mean": ["mean value of X_EDIST", "Float"],
    "X_LENGTH_mean": ["mean value of X_LENGTH", "Float"],
    "X_MAPQ_mean": ["mean value of X_MAPQ", "Float"],
    "X_FC1_mean": ["mean value of X_FC1", "Float"],
    "X_FC2_mean": ["mean value of X_FC2", "Float"],
    "MAX_SOFTCLIP_LENGTH_mean": ["mean value of MAX_SOFTCLIP_LENGTH", "Float"],
    "X_FLAGS_mean": ["mean value of X_FLAGS", "Float"],
    "ML_QUAL_mean": ["mean value of ML_QUAL", "Float"],
    "X_QUAL_max": ["max value of X_QUAL", "Float"],
    "X_INDEX_max": ["max value of X_INDEX", "Integer"],
    "X_QUAL_min": ["min value of X_QUAL", "Float"],
    "X_INDEX_min": ["min value of X_INDEX", "Integer"],
    "count_forward": ["number of forward reads", "Integer"],
    "count_reverse": ["number of reverse reads", "Integer"],
    "count_duplicate": ["number of duplicate reads", "Integer"],
    "count_non_duplicate": ["number of non-duplicate reads", "Integer"],
}

ppm_added_agg_features = {}
ppm_added_agg_features["st_MINUS"] = ["number of st tagged as MINUS", "Integer"]
ppm_added_agg_features["st_MIXED"] = ["number of st tagged as MIXED", "Integer"]
ppm_added_agg_features["st_PLUS"] = ["number of st tagged as PLUS", "Integer"]
ppm_added_agg_features["st_UNDETERMINED"] = ["number of st tagged as UNDETERMINED", "Integer"]
ppm_added_agg_features["et_MINUS"] = ["number of et tagged as MINUS", "Integer"]
ppm_added_agg_features["et_MIXED"] = ["number of et tagged as MIXED", "Integer"]
ppm_added_agg_features["et_PLUS"] = ["number of et tagged as PLUS", "Integer"]
ppm_added_agg_features["et_UNDETERMINED"] = ["number of et tagged as UNDETERMINED", "Integer"]
ppm_added_agg_features["num_mixed_reads"] = ["number of mixed reads", "Integer"]


def record_manual_aggregation(rec, xgb_model=None):  # noqa: C901
    record_info_dict = dict(rec.info)
    record_dict_for_xgb = {}

    # add non aggrgate fields
    record_dict_for_xgb["alt_reads"] = rec.samples[0]["AD"][1]
    record_dict_for_xgb["ref_allele"] = rec.alleles[0]
    record_dict_for_xgb["alt_allele"] = rec.alleles[1]
    record_dict_for_xgb["ref"] = rec.ref
    record_dict_for_xgb["qual"] = rec.qual
    non_agg_fields_for_xgb_from_rec = ["DP", "VAF"]
    for field in non_agg_fields_for_xgb_from_rec:
        record_dict_for_xgb[field] = rec.samples[0][field]
    non_agg_fields_for_xgb_from_rec_dict = featuremap_consensus_utils.fields_to_collect_all_options[
        "fields_to_write_once"
    ]
    for field in non_agg_fields_for_xgb_from_rec_dict:
        record_dict_for_xgb[field] = record_info_dict[field]

    # add aggregate fields
    for colname in columns_for_mean_aggregation:
        agg_colname = colname.upper() + "_mean"
        record_dict_for_xgb[agg_colname] = statistics.mean(record_info_dict[colname])
    for colname in columns_for_min_aggregation:
        agg_colname = colname.upper() + "_min"
        record_dict_for_xgb[agg_colname] = min(record_info_dict[colname])
    for colname in columns_for_max_aggregation:
        agg_colname = colname.upper() + "_max"
        record_dict_for_xgb[agg_colname] = max(record_info_dict[colname])
    record_dict_for_xgb["count_forward"] = record_info_dict["is_forward"].split("|").count("T")
    record_dict_for_xgb["count_reverse"] = record_info_dict["is_forward"].split("|").count("F")
    record_dict_for_xgb["count_duplicate"] = record_info_dict["is_duplicate"].split("|").count(True)
    record_dict_for_xgb["count_non_duplicate"] = record_info_dict["is_duplicate"].split("|").count(False)
    if "is_cycle_skip" in record_info_dict:
        record_dict_for_xgb["is_cycle_skip"] = record_info_dict["is_cycle_skip"]
    else:
        record_dict_for_xgb["is_cycle_skip"] = False

    if "st" in record_info_dict:  # ppmseq columns
        record_dict_for_xgb["num_mixed_reads"] = sum(
            [
                1
                for i, j in zip(record_info_dict["st"], record_info_dict["et"], strict=False)
                if (i == j) & (i == "MIXED")
            ]
        )
        for colname in columns_for_st_et_aggregation:
            tags = ["MINUS", "MIXED", "PLUS", "UNDETERMINED"]
            for tag in tags:
                record_dict_for_xgb[colname + "_" + tag] = record_info_dict[colname].split("|").count(tag)

    if xgb_model:
        record_dict_for_xgb["xgb_proba"] = predict_record_with_xgb(record_dict_for_xgb, xgb_model)

    return record_dict_for_xgb


def set_categorial_columns(df):
    categorical_columns = df.select_dtypes(include=["object", "category"]).columns
    # print(categorical_columns)
    le = LabelEncoder()
    for col in categorical_columns:
        df.loc[:, col] = le.fit_transform(df[col].astype(str))


def predict_record_with_xgb(record_dict_for_xgb, xgb_model):
    """
    Predict the record with the xgb model
    Inputs:
        record_dict_for_xgb (dict): A dictionary containing the record information
        xgb_model (xgboost.XGBClassifier): The xgboost model
    Output:
        probability_value (float): The xgb probability of the record to be a true variant
    """

    # load xgb model
    xgb_clf_es = xgboost.XGBClassifier()
    xgb_clf_es.load_model(xgb_model)
    features = xgb_clf_es.get_booster().feature_names
    # prepare dataframe
    xgb_df = pd.DataFrame(record_dict_for_xgb, index=[0])
    xgb_df = xgb_df.rename(
        columns={
            "DP": "dp",
            "X_READ_COUNT": "x_read_count",
            "X_FILTERED_COUNT": "x_filtered_count",
            "count_forward": "num_is_forward",
            "VAF": "vaf",
        }
    )

    X = xgb_df[features]  # noqa: N806
    set_categorial_columns(X)
    X = X.fillna(0)  # noqa: N806

    # predict record
    probabilities = xgb_clf_es.predict_proba(X)
    df_probabilities = pd.DataFrame(probabilities, columns=["0", "1"])
    return df_probabilities["1"].to_numpy()[0]


def pileup_featuremap_with_agg_params_and_xgb_proba(  # noqa: C901
    featuremap_vcf_file: str,
    output_vcf: str,
    genomic_interval: str = None,
    xgb_model: xgboost.XGBClassifier = None,
    verbose: bool = True,  # noqa: FBT001, FBT002
) -> str:
    """
    Write the vcf file with the aggregated fields and the xgb probability
    """
    if verbose:
        logger.setLevel(logging.DEBUG)
        for handler in logger.handlers:
            handler.setLevel(logging.DEBUG)

    with tempfile.TemporaryDirectory(dir=dirname(output_vcf)) as temp_dir:
        sorted_featuremap = pjoin(temp_dir, basename(featuremap_vcf_file).replace(".vcf.gz", ".sorted.vcf.gz"))
        if genomic_interval is None:
            # sort all featuremap
            sort_cmd = f"bcftools sort {featuremap_vcf_file} -Oz -o {sorted_featuremap} && bcftools index -t {sorted_featuremap}"  # noqa: E501
        else:
            # sort only the genomic interval of interest
            sort_cmd = f"bcftools view {featuremap_vcf_file} {genomic_interval} |\
                    bcftools sort - -Oz -o {sorted_featuremap} && bcftools index -t {sorted_featuremap}"
        logger.debug(sort_cmd)
        subprocess.check_call(sort_cmd, shell=True)  # noqa: S602

        with pysam.VariantFile(sorted_featuremap) as vcfin:
            hdr = vcfin.header
            # adding manual aggregation fields
            # for field, field_type, field_description in zip(
            for field in added_agg_features:
                field_type = added_agg_features[field][1]
                field_description = added_agg_features[field][0]
                hdr.info.add(field, 1, field_type, field_description)
            if "st" in hdr.info:
                for field in ppm_added_agg_features:
                    field_type = ppm_added_agg_features[field][1]
                    field_description = ppm_added_agg_features[field][0]
                    hdr.info.add(field, 1, field_type, field_description)
            hdr.info.add("xgb_proba", 1, "Float", "XGBoost probability of the record to be a true variant")
            with pysam.VariantFile(output_vcf, mode="w", header=hdr) as vcfout:
                for row in vcfin:
                    record_dict_for_xgb = record_manual_aggregation(row, xgb_model)
                    for key in added_agg_features:
                        row.info[key] = record_dict_for_xgb[key]
                    if "st" in hdr.info:
                        for key in ppm_added_agg_features:
                            row.info[key] = record_dict_for_xgb[key]
                    row.info["xgb_proba"] = float(record_dict_for_xgb["xgb_proba"])
                    vcfout.write(row)
            vcfout.close()
            vcfin.close()
    pysam.tabix_index(output_vcf, preset="vcf", min_shift=0, force=True)
    return output_vcf


def pileup_featuremap_with_agg_params_and_xgb_proba_on_an_interval_list(
    featuremap_vcf_file: str,
    output_vcf: str,
    interval_list: str,
    xgb_model: str = None,
    verbose: bool = True,  # noqa: FBT001, FBT002
) -> None:
    """
    Apply pileup featuremap on an interval list
    Inputs:
        featuremap (str): The input featuremap vcf file
        output_vcf (str): The output pileup vcf file
        interval_list (str): The interval list file
        xgb_model (str): The xgboost model file
        verbose (bool): The verbosity level (default: True)
    Output:
        output_vcf (str): The output vcf file including the aggregated fields and the xgb probability
    """
    if not output_vcf.endswith(".vcf.gz"):
        logger.debug("adding .vcf.gz suffix to the output vcf file")
        output_vcf = output_vcf + ".vcf.gz"

    with tempfile.TemporaryDirectory(dir=dirname(output_vcf)):
        genomic_intervals = []
        with open(interval_list, encoding="utf-8") as f:
            for line in f:
                # ignore header lines
                if line.startswith("@"):
                    continue
                # read genomic ineterval
                genomic_interval = line.strip()
                genomic_interval_list = genomic_interval.split("\t")
                chrom = genomic_interval_list[0]
                start = genomic_interval_list[1]
                end = genomic_interval_list[2]
                genomic_interval = chrom + ":" + str(start) + "-" + str(end)
                genomic_intervals.append(genomic_interval)

        params = [
            (
                featuremap_vcf_file,
                f"{output_vcf}.{genomic_interval}.int_list.vcf.gz",
                genomic_interval,
                xgb_model,
                verbose,
            )
            for genomic_interval in genomic_intervals
        ]
        num_cpus = os.cpu_count()
        with ThreadPoolExecutor(max_workers=num_cpus) as executor:
            results = list(executor.map(lambda p: pileup_featuremap_with_agg_params_and_xgb_proba(*p), params))

        # merge the output vcfs
        vcf_str = " ".join(results)
        cmd = f"bcftools concat {vcf_str} -a | bcftools sort - -Oz -o {output_vcf} && bcftools index -t {output_vcf}"
        logger.debug(cmd)
        subprocess.check_call(cmd, shell=True)  # noqa: S602
    return output_vcf


def __parse_args(argv: list[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="add_aggregate_parameters_and_xgb_score_to_pileup_featuremap",
        description=run.__doc__,
    )
    parser.add_argument(
        "-f",
        "--featuremap_pileup",
        type=str,
        required=True,
        help="""Featuremap pileup vcf file""",
    )
    parser.add_argument(
        "-o",
        "--output_vcf",
        type=str,
        required=True,
        help="""Output pileup vcf file""",
    )
    parser.add_argument(
        "-i",
        "--interval_list_file",
        type=str,
        required=True,
        help="""Interval list file""",
    )
    parser.add_argument(
        "-m",
        "--xgb_model_file",
        type=str,
        required=True,
        help="""XGBoost model file""",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        type=bool,
        required=False,
        default=True,
        help="""Whether to print debug messages (default: True)""",
    )
    return parser.parse_args(argv[1:])


def run(argv):
    """Add aggregated parameters and xgb probability to the featuremap pileup vcf file"""
    args_in = __parse_args(argv)
    pileup_featuremap_with_agg_params_and_xgb_proba_on_an_interval_list(
        args_in.featuremap_pileup,
        args_in.output_vcf,
        args_in.interval_list_file,
        args_in.xgb_model_file,
        args_in.verbose,
    )


def main():
    run(sys.argv)


if __name__ == "__main__":
    main()



import argparse
import logging
import sys
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pysam
import xgboost as xgb
from sklearn.metrics import accuracy_score, classification_report
from sklearn.model_selection import cross_val_score, train_test_split
from ugbio_core.logger import logger
from ugbio_core.vcfbed import vcftools

from ugbio_featuremap import featuremap_xgb_prediction
from ugbio_featuremap.featuremap_utils import FeatureMapFields

added_agg_features = featuremap_xgb_prediction.added_agg_features
ppm_added_agg_features = featuremap_xgb_prediction.ppm_added_agg_features
custom_info_fields = [
    FeatureMapFields.READ_COUNT.value,
    FeatureMapFields.FILTERED_COUNT.value,
    FeatureMapFields.TRINUC_CONTEXT_WITH_ALT.value,
    FeatureMapFields.HMER_CONTEXT_REF.value,
    FeatureMapFields.HMER_CONTEXT_ALT.value,
    FeatureMapFields.PREV_1.value,
    FeatureMapFields.PREV_2.value,
    FeatureMapFields.PREV_3.value,
    FeatureMapFields.NEXT_1.value,
    FeatureMapFields.NEXT_2.value,
    FeatureMapFields.NEXT_3.value,
    FeatureMapFields.IS_CYCLE_SKIP.value,
]
custom_info_fields.extend(list(added_agg_features))


def split_data_every_2nd_variant(X, y):  # noqa: N802, N803
    # Split the data into training and testing sets - every 2nd record
    X_train = X.iloc[::2]  # Select rows with even indices  # noqa: N806
    X_test = X.iloc[1::2]  # Select rows with odd indices  # noqa: N806
    y_train = y.iloc[::2]
    y_test = y.iloc[1::2]
    return [X_train, X_test, y_train, y_test]


def split_data(X, y, test_size=0.25, random_state=42):  # noqa: N802, N803
    # Split the data into training and testing sets
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size, random_state=random_state)  # noqa: N806
    return [X_train, X_test, y_train, y_test]


def XGBoost_train(X_train, y_train):  # noqa: N802, N803
    # Create XGBoost classifier
    xgb_clf_es = xgb.XGBClassifier(
        objective="binary:logistic",
        n_estimators=1000,  # Set a high number, early stopping will decide when to stop
        learning_rate=0.1,
        max_depth=3,
        random_state=42,
        enable_categorical=True,
    )
    # Fit the model
    xgb_clf_es.fit(X_train, y_train, verbose=True)
    return xgb_clf_es


def XGBoost_test(xgb_clf_es, X_test):  # noqa: N802, N803
    # Make predictions
    y_pred_es = xgb_clf_es.predict(X_test)
    return y_pred_es


def XGBoost_evaluate_model(xgb_clf_es, y_test, y_pred_es):  # noqa: N802, N803
    # Evaluate the model
    accuracy_es = accuracy_score(y_test, y_pred_es)
    logger.debug(f"\nAccuracy : {accuracy_es:.2f}")

    logger.debug(f"Accuracy:{accuracy_score(y_test, y_pred_es)}")
    logger.debug(f"\nClassification Report:\n{classification_report(y_test, y_pred_es)}")

    # Feature importance
    feature_importance = pd.DataFrame(
        {"feature": xgb_clf_es.get_booster().feature_names, "importance": xgb_clf_es.feature_importances_}
    ).sort_values("importance", ascending=False)
    logger.debug("\nFeature Importance:")
    logger.debug(f"{feature_importance}")


def variants_classification_vs_probability(xgb_clf_es, X_test, y_test, out_figure_path):  # noqa: N802, N803
    probabilities = xgb_clf_es.predict_proba(X_test)
    df_probabilities = pd.DataFrame(probabilities, columns=["0", "1"])
    df_probabilities["true_status"] = pd.Series(y_test.values)

    plt.figure(figsize=(10, 6))
    plt.hist(df_probabilities[df_probabilities["true_status"] == 1]["1"], bins=20, alpha=0.4, label="True Variants")
    plt.hist(df_probabilities[df_probabilities["true_status"] == 0]["1"], bins=20, alpha=0.4, label="False Variants")

    # Add labels and legend
    plt.xlabel("probability")
    plt.ylabel("Frequency")
    plt.legend()
    plt.yscale("log")
    plt.savefig(out_figure_path)


def cross_validation(xgb_clf_es, X, y, cv=5, cv_training_diff_cutoff=0.1):  # noqa: N802, N803
    cv_scores = cross_val_score(xgb_clf_es, X, y, cv=cv)

    logger.debug(f"Cross-validation scores: {cv_scores}")
    logger.debug(f"Mean CV score: {np.mean(cv_scores):.4f}")
    logger.debug(f"Standard deviation of CV scores: {np.std(cv_scores):.4f}")

    # If CV scores are significantly lower than training score, it may indicate overfitting
    if xgb_clf_es.score(X, y) - np.mean(cv_scores) > cv_training_diff_cutoff:  # This threshold can be adjusted
        logger.debug("Model might be overfitting.")
    else:
        logger.debug("No clear sign of overfitting.")


def aggregate_vcf(sorted_featuremap: str, output_vcf: str, added_agg_features: dict, ppm_added_agg_features: dict):
    with pysam.VariantFile(sorted_featuremap) as vcfin:
        ### add new aggregation fields to header
        hdr = vcfin.header
        for field in added_agg_features:
            field_type = added_agg_features[field][1]
            field_description = added_agg_features[field][0]
            hdr.info.add(field, 1, field_type, field_description)
        if "st" in hdr.info:  # for ppmSeq
            for field in ppm_added_agg_features:
                field_type = ppm_added_agg_features[field][1]
                field_description = ppm_added_agg_features[field][0]
                hdr.info.add(field, 1, field_type, field_description)
        ### calculate aggregation fields per record and write to new out vcf
        with pysam.VariantFile(output_vcf, mode="w", header=hdr) as vcfout:
            for row in vcfin:
                record_dict_for_xgb = featuremap_xgb_prediction.record_manual_aggregation(row)
                for key in added_agg_features:
                    row.info[key] = record_dict_for_xgb[key]
                if "st" in hdr.info:
                    for key in ppm_added_agg_features:
                        row.info[key] = record_dict_for_xgb[key]
                vcfout.write(row)
        vcfout.close()
        vcfin.close()
    pysam.tabix_index(output_vcf, preset="vcf", min_shift=0, force=True)
    return output_vcf


def __parse_args(argv: list[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="train xgboost model using pileup featuremap aggregated data",
        description=run.__doc__,
    )
    parser.add_argument(
        "-tp",
        "--featuremap_pileup_tp",
        type=str,
        required=True,
        help="""Featuremap pileup vcf file holding true positive variants""",
    )
    parser.add_argument(
        "-fp",
        "--featuremap_pileup_fp",
        type=str,
        required=True,
        help="""Featuremap pileup vcf file holding false positive variants""",
    )
    parser.add_argument(
        "-o",
        "--output_model",
        type=str,
        required=True,
        help="""Output model file""",
    )
    parser.add_argument(
        "-min_alt_reads",
        "--min_supporting_reads_cutoff",
        type=str,
        required=True,
        help="""minimal value of alternative supporting reads""",
    )
    parser.add_argument(
        "-max_alt_reads",
        "--max_supporting_reads_cutoff",
        type=str,
        required=True,
        help="""maximal value of alternative supporting reads""",
    )
    parser.add_argument(
        "-chr",
        "--chromosome",
        type=str,
        required=False,
        help="""Optional chromosome name to use for data loading""",
    )
    parser.add_argument(
        "-is_ppm",
        "--is_ppmSeq",
        action="store_true",
        help="""Wether the input featuremap_pileup is ppmeSeq""",
    )
    parser.add_argument(
        "-split_data_every_2nd_variant",
        "--split_data_every_2nd_variant",
        action="store_true",
        help=(
            "Wether to split the data every 2nd variant (resulting with 50% training and 50% testing). "
            "Overrides the default split_data function"
        ),
    )
    parser.add_argument(
        "-test_size",
        "--test_size",
        type=float,
        default=0.25,
        required=False,
        help=(
            "Optional test size fraction for the split_data function. "
            "Default is 0.25 (25% for testing and 75% for training)"
        ),
    )
    parser.add_argument(
        "-v",
        "--verbose",
        type=bool,
        required=False,
        default=True,
        help="""Whether to print debug messages (default: True)""",
    )
    return parser.parse_args(argv[1:])


def run(argv):  # noqa: C901,PLR0912,PLR0915
    """train xgboost model using pileup featuremap aggregated data"""
    args_in = __parse_args(argv)

    if args_in.verbose:
        logger.setLevel(logging.DEBUG)
        for handler in logger.handlers:
            handler.setLevel(logging.DEBUG)

    out_file = Path(args_in.output_model)
    if out_file.is_file():
        logger.debug(f"out file exists: {out_file}. please provide a different output file path")
        sys.exit(1)

    out_file = args_in.output_model
    if not out_file.endswith(".json"):
        logger.debug("adding .json suffix to the output vcf file")
        out_file = out_file + ".json"

    if args_in.is_ppmSeq:
        custom_info_fields.extend(list(ppm_added_agg_features))

    training_features = ["dp", "vaf", "chrom", "pos", "qual", "ref"]
    training_features.extend(custom_info_fields)
    training_features = [f.lower() for f in training_features]

    # aggregate vcf
    vcf_agg_file = {}
    for tag, sorted_featuremap in zip(
        ["tp", "fp"], [args_in.featuremap_pileup_tp, args_in.featuremap_pileup_fp], strict=False
    ):
        output_vcf = sorted_featuremap.replace(".vcf.gz", ".agg_params.vcf.gz")
        output_vcf = aggregate_vcf(sorted_featuremap, output_vcf, added_agg_features, ppm_added_agg_features)
        vcf_agg_file[tag] = output_vcf

    logger.debug("finished writing aggregate params vcf files:")
    logger.debug(vcf_agg_file["fp"])
    logger.debug(vcf_agg_file["tp"])

    df_fp = vcftools.get_vcf_df(
        vcf_agg_file["fp"], custom_info_fields=custom_info_fields, chromosome=args_in.chromosome
    )
    df_tp = vcftools.get_vcf_df(
        vcf_agg_file["tp"], custom_info_fields=custom_info_fields, chromosome=args_in.chromosome
    )
    df_fp["label"] = "negative"
    df_tp["label"] = "positive"
    df_all_variants = pd.concat([df_fp, df_tp])
    logger.debug("finished reading pileup featurepmap vcf files")

    lower_cutoff = int(args_in.min_supporting_reads_cutoff)  # noqa: F841
    upper_cutoff = int(args_in.max_supporting_reads_cutoff)  # noqa: F841

    X = df_all_variants.query("alt_reads>=@lower_cutoff and alt_reads<=@upper_cutoff")[training_features]  # noqa: N806
    y = df_all_variants.query("alt_reads>=@lower_cutoff and alt_reads<=@upper_cutoff")["label"].apply(
        lambda x: 1 if x == "positive" else 0
    )

    featuremap_xgb_prediction.set_categorial_columns(X)
    for col in X.select_dtypes(include="object").columns:
        X[col] = X[col].astype("category")

    if args_in.split_data_every_2nd_variant:
        [X_train, X_test, y_train, y_test] = split_data_every_2nd_variant(X, y)  # noqa: N806
    elif args_in.test_size < 1 and args_in.test_size > 0:
        [X_train, X_test, y_train, y_test] = split_data(X, y, test_size=args_in.test_size)  # noqa: N806
    xgb_clf_es = XGBoost_train(X_train, y_train)

    logger.debug("Model evaluation on TEST data:")
    y_pred_es = XGBoost_test(xgb_clf_es, X_test)
    XGBoost_evaluate_model(xgb_clf_es, y_test, y_pred_es)

    out_figure_path = out_file.replace(".json", ".probability_histogram.png")
    variants_classification_vs_probability(xgb_clf_es, X_test, y_test, out_figure_path)

    logger.debug("Model evaluation on TRAIN data")
    y_pred_es = XGBoost_test(xgb_clf_es, X_train)
    XGBoost_evaluate_model(xgb_clf_es, y_train, y_pred_es)

    cross_validation(xgb_clf_es, X, y)

    # save model
    xgb_clf_es.save_model(out_file)
    logger.debug(f"out model is save in: {out_file}")


def main():
    run(sys.argv)


if __name__ == "__main__":
    main()
